/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */

/**********************************************************************
  node  -- represents frames in Parka
  **********************************************************************/
package parka;

import java.awt.*;
import java.util.*;

public class Node extends parka.NodeObject
{
    // all nodes in this graph
    //
    static  Vector  nodesVector = new Vector();
  
    // set of links associated with this node
    private Vector linkslist;

    // set of edges out of this node
    private Vector edgelist;

    // color
    final static Color HI_COLOR = Color.yellow;

    // The background color for the Node - changes when un/selected.
    private Color c;
    private static  int  nextInt = 0;

    // Constants
    final static int XBUF = 6;
    final static int YBUF = 2;

    public final static char  VARIABLE_PREFIX = '?';

    // is this a variable or constant
    private boolean isvariable;

    // The binding can be "".
    private String binding;
    private String nodeLabel;
	//when binding is not null, it is combined with superclass's 'str'
        //(the object's name) and the binding, here.  Used for display only.

    // *******
    // Methods
    // *******

    Node (DrawArea MyDraw, int x, int y)
    {
        super (MyDraw, x, y);

        isvariable = true;    // variable by default
        binding = null;       // no binding initially
        nodeLabel = null;

        str = "noname" + String.valueOf (nextInt);
	nextInt++;
	if (nextInt == 1000)   //assume we never need this many
            nextInt = 0;
        width = resetWidth();

        linkslist = new Vector();
        edgelist = new Vector();

	// add to global list of nodes:
        Node.nodesVector.addElement (this);

        c = DEF_COLOR;      //background color of node
	//parent class has variable 'draw' to remember DrawArea its in
    }

    // Do in the reverse order that they were drawn so you select
    // top objects first
    //
    static Node hitTest (int x, int y)
    {
	for (int i=(nodesVector.size()-1); i >= 0; i--)
        {
            Node   obj = (Node) nodesVector.elementAt(i);

            if (obj.mouseWithin(x,y))
                return obj;
        }
	return  null;
    }

    boolean mouseWithin (int x, int y)
    {
	int  xchange = width/2 + XBUF;
	int  ychange = height/2 + YBUF;

        return ((x > X - xchange) && (x < X + xchange) &&
                (y > Y - ychange) && (y < Y + ychange));
    }

    static void drawNodes (Graphics g)
    {
        Enumeration e;
	Node        node;

        e = Node.nodesVector.elements();
        while (e.hasMoreElements())
        {
            try
            {
                node = (Node)e.nextElement();
                node.drawGO (g);
            }
            catch (NoSuchElementException e1)
            {
                // Should not get here.
		System.out.println ("Exception: No such frame node");
            }
        }
    }

    // Draw the GraphObject
    void drawGO (Graphics g)
    {
	int  xoff = draw.getXSBoffset();
	int  yoff = draw.getYSBoffset();

        // draw the base of the node
        g.setColor(c);

	//positions:
        int  top = X - width/2 - XBUF - xoff;
	int  left = Y - height/2 - YBUF - yoff;
	int  localwidth = width + XBUF*2;
	int  localheight = height + YBUF*2;

        // if it's a variable, use a round-rect
        if (isvariable)
        {
            g.fillRoundRect (top, left, localwidth, localheight,
                                       height, height);
            g.setColor (Color.black);
            g.drawRoundRect (top, left, localwidth, localheight,
                                       height, height);
        }
        else
        {
            g.fillRect (top, left, localwidth, localheight);
            g.setColor (Color.black);
            g.drawRect (top, left, localwidth, localheight);
        }

	//System.out.println ("drawing label string now; width = " + width);
        g.drawString (getObjLabel(), X - width/2 - xoff,
                       Y + height/2 - descent - yoff);
    }

    // move this Node to first position in node vector, so that it
    // pops to the top if covered
    //
    void raiseSelected()
    {
        nodesVector.removeElement (this);
        nodesVector.addElement (this);

        // move this Node's Links to the top too.
        Enumeration e = linkslist.elements();
        while (e.hasMoreElements())
        {
            Link ln = (Link)e.nextElement();
            ln.raiseSelected();
        }
    }

    static void updateFont (DrawArea da)
    {
	NodeObject.updateFont(da);

	Enumeration elts = nodesVector.elements();
        while (elts.hasMoreElements())
        {
            Node n = (Node)elts.nextElement();
	    n.width = n.resetWidth();
	    n.height = n.resetHeight();
	}
    }

    // Add Link
    void addLink (Link ln, Edge e)
    {
        linkslist.addElement (ln);
	edgelist.addElement (e);
    }

    // Delete link from this node's memory
    void deleteLink(Link ln)
    {
        linkslist.removeElement(ln);
    }

    // Delete this node
    void deleteGO()
    {
	removeAllEdges();
        nodesVector.removeElement(this);
    }

    //should only be called from Node or Link's 'deleteEdge routine.
    protected void delete (Edge e, Link otherEnd)
    {
	for (int i = 0; i < linkslist.size(); i++)
        {
	    if (otherEnd == linkslist.elementAt(i))
	    {
		linkslist.removeElementAt(i);   //do NOT delete
		break;
	    }
        }
	for (int i = 0; i < edgelist.size(); i++)
        {
	    if (e == (Edge)edgelist.elementAt(i))
	    {
		edgelist.removeElementAt(i);
		e.deleteNode();
		break;
	    }
	}
    }

    protected void deleteEdge (Edge e)
    {
        Link  otherEnd = e.getLink();

        if (otherEnd != null)
	{
            delete (e, otherEnd);
            otherEnd.delete (e, this);
	    Edge.edgesVector.removeElement (e);
	}
    }

    //removes Edge from edgelist and deletes it;
    //removes Link from linkslist
    //
    void removeAllEdges()
    {
	// here we use a clone of the Vector (which will be no problem,
	// because it's a shallow copy) in order to delete from the
	// edgelist without screwing up the internal pointers to items
	// when Enumerating
	//
        Vector  edgelistCopy = (Vector)edgelist.clone();
	Enumeration elts = edgelistCopy.elements();
        while (elts.hasMoreElements())
        {
            Edge e = (Edge)elts.nextElement();
            deleteEdge (e);
        }
    }

    // For selecting the GraphObject
    void selectGO()
    {
	//System.out.println ("Node.selectGO()");

        c = HI_COLOR;
	active = this;
	raiseSelected();

        // set text field new node's name; set "variable" button
	// display status
	draw.nodeSelectionMode (this);

        // Change menus
	draw.toggleMenuStatus (true, getIsVariable());
	     //obj. menu becomes sensitive; link menu is made insensitive
    }

    // For deselecting the GraphObject
    void unselectGO()
    {
	//System.out.println ("Node.unselectGO()");

        c = DEF_COLOR;
	active = null;
    }

    // Rename the Node (Change the string in the object)
    void renameGO (String newstr)
    {
	setBindingAppearance (newstr);
	// Update the Node size
        str = newstr;
        width = resetWidth();
        height = resetHeight();

        // System.out.println ("width changed (rename):  " + width);
    }

    // Variable/Constant status can toggle
    void  changeVariable (boolean val, boolean togglep)
    {
        if (togglep)
            isvariable = !isvariable;
        else
            isvariable = val;
    }

    boolean getIsVariable()
    {
	return  isvariable;
    }

    // Return the variable with the give name
    //
    public static Node getVariable (String str)
    {
	// System.out.println ("       " + str);

        String   strWithoutPrefix;
        Character first = new Character (str.charAt(0));
        Character oneToMatch = new Character (Node.VARIABLE_PREFIX);
	if (first.equals (oneToMatch))
	    strWithoutPrefix = str.substring(1);
	else
	    strWithoutPrefix = str;

	// System.out.println ("       " + strWithoutPrefix);

        Enumeration e = nodesVector.elements();
        Node     node;    
        while (e.hasMoreElements())
        {
            node = (Node)e.nextElement();
	    // System.out.println ("       this node's name: <"
            //              + node.getObjName() + ">");

            if ((node.getObjName()).equals (strWithoutPrefix))
                return  node;
        }

        return null;
     }

    static boolean changeText (String str, DrawArea da)
    {
        if (active == null)
            return  false;

	String  trimmedStr = str.trim();
	if (trimmedStr.length() == 0)  //then only blanks were there
	    return  false;             //don't change name

        // Change Node name
	if (active instanceof Node)
            ((Node)active).renameGO(str);

	// TO DO:   Re: MultiColumn object:
        // change column title if active is a variable and resultArea exists
	// A registration/callback system, to decouple from graph?
        //if (active.getIsVariable())

        da.paintDrawArea();

        return true;
    }

    // Remove all bindings from the query graph
    static void unbindVariables()
    {
        Enumeration e = nodesVector.elements();
        while (e.hasMoreElements()) 
            ((Node)e.nextElement()).changeBinding(null);
    }

    // overrides parent's routine
    String getObjLabel()
    {
	if (nodeLabel == null)
	{
	    // System.out.println ("object label is null -- returns [" + str +
                      //          "]; width = " + width);
	    return  str;
	}
	else
	{
	    // System.out.println ("returns object label [" + nodeLabel +
                      //          "]; width = " + width);
	    return  nodeLabel;
	}
    }

    // returns variable name without bindings, if there are any
    // while bindings to the variable exist, the super class's 'str'
    // includes the var name and the binding.  In this class, we
    // remember the original var name in 'origVarName'.  When the
    // bindings are removed, 'origVarName' is made null, and 'str' is
    // restored.
    private void setBindingAppearance (String obname)
    {      
        if (isvariable && (binding != null) && (nodeLabel == null))
        {
            nodeLabel = obname + ": " + binding;
        }
        else  if (isvariable && (binding == null) && (nodeLabel != null))
        {
	    nodeLabel = null;
        }
    }

    //set binding value
    void changeBinding (String newstr)
    {
        binding = newstr;
	setBindingAppearance (str);
	width = resetWidth();
    }

    //get binding value
    String getBinding ()
    {
	return binding;
    }

    static void clearAll()
    {
        nodesVector.removeAllElements();
    }

    public void debug (String str)
    {
        System.out.println("");
        System.out.println(str + "Node: " + this.str);
        System.out.println(str + "Node label: " + nodeLabel);
        System.out.println(str + " " + linkslist.size() + " links");
        for (int i=0; i<linkslist.size(); i++) {
            Link ln = (Link)linkslist.elementAt(i);
            ln.debug(str + "  ");
        }
    }

    public static void printDebug()
    {
	System.out.println("");
        System.out.println (nodesVector.size() + " nodes");
        for (int i = 0; i < nodesVector.size(); i++)
        {
            Node n = (Node)nodesVector.elementAt(i);
            n.debug(" ");
        }
    }
}
